/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2008 Stian Selnes <<user@hostname.org>>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-suppresscolor
 *
 * <refsect2>
 * <title>Example launch line</title>
 * <para>
 * <programlisting>
 * gst-launch -v -m videotestsrc ! suppresscolor ! fakesink silent=TRUE
 * </programlisting>
 * </para>
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/video/video.h>

#include "gstsuppresscolor.h"

#include <string.h>

GST_DEBUG_CATEGORY_STATIC (gst_suppresscolor_debug);
#define GST_CAT_DEFAULT gst_suppresscolor_debug


enum
{
	/* FILL ME */
	LAST_SIGNAL
};

enum
{
	PROP_0,
	PROP_SUPPRESS_Y,
	PROP_SUPPRESS_U,
	PROP_SUPPRESS_V,
	PROP_SUPPRESSION_VALUE_Y,
	PROP_SUPPRESSION_VALUE_U,
	PROP_SUPPRESSION_VALUE_V
};

#define DEFAULT_SUPPRESSION_VALUE_Y 127
#define DEFAULT_SUPPRESSION_VALUE_U 127
#define DEFAULT_SUPPRESSION_VALUE_V 127


static const GstElementDetails suppresscolor_details =
GST_ELEMENT_DETAILS( "Suppress color component(s)",
                     "Filter/Video",
                     "Suppress given color component(s) in video stream",
                     "Stian Selnes <<stian.selnes@gmail.com>>");

static GstStaticPadTemplate
gst_suppresscolor_sink_template = GST_STATIC_PAD_TEMPLATE ("sink",
                                                           GST_PAD_SINK,
                                                           GST_PAD_ALWAYS,
                                                           GST_STATIC_CAPS (GST_VIDEO_CAPS_YUV ("I420")));

static GstStaticPadTemplate
gst_suppresscolor_src_template = GST_STATIC_PAD_TEMPLATE ("src",
                                                          GST_PAD_SRC,
                                                          GST_PAD_ALWAYS,
                                                          GST_STATIC_CAPS (GST_VIDEO_CAPS_YUV ("I420")));

GST_BOILERPLATE (GstSuppressColor, gst_suppresscolor, GstVideoFilter, GST_TYPE_VIDEO_FILTER);

static void gst_suppresscolor_set_property (GObject * object, guint prop_id,
                                            const GValue * value, GParamSpec * pspec);
static void gst_suppresscolor_get_property (GObject * object, guint prop_id,
                                            GValue * value, GParamSpec * pspec);
static gboolean gst_suppresscolor_set_caps (GstBaseTransform *trans, GstCaps *incaps,
                                            GstCaps *outcaps);
static GstFlowReturn gst_suppresscolor_transform_ip (GstBaseTransform * transform,
                                                     GstBuffer * buf);

/* Macros borrowed from gamma plugin */
#define GST_VIDEO_I420_Y_ROWSTRIDE(width) (GST_ROUND_UP_4(width))
#define GST_VIDEO_I420_U_ROWSTRIDE(width) (GST_ROUND_UP_8(width)/2)
#define GST_VIDEO_I420_V_ROWSTRIDE(width) ((GST_ROUND_UP_8(GST_VIDEO_I420_Y_ROWSTRIDE(width)))/2)

#define GST_VIDEO_I420_Y_OFFSET(w,h) (0)
#define GST_VIDEO_I420_U_OFFSET(w,h) (GST_VIDEO_I420_Y_OFFSET(w,h)+(GST_VIDEO_I420_Y_ROWSTRIDE(w)*GST_ROUND_UP_2(h)))
#define GST_VIDEO_I420_V_OFFSET(w,h) (GST_VIDEO_I420_U_OFFSET(w,h)+(GST_VIDEO_I420_U_ROWSTRIDE(w)*GST_ROUND_UP_2(h)/2))
#define GST_VIDEO_I420_SIZE(w,h)     (GST_VIDEO_I420_V_OFFSET(w,h)+(GST_VIDEO_I420_V_ROWSTRIDE(w)*GST_ROUND_UP_2(h)/2))


static void
gst_suppresscolor_base_init (gpointer gclass)
{
	GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

	gst_element_class_add_pad_template (
		element_class,
		gst_static_pad_template_get (&gst_suppresscolor_src_template));

	gst_element_class_add_pad_template (
		element_class,
		gst_static_pad_template_get (&gst_suppresscolor_sink_template));

	gst_element_class_set_details (element_class, &suppresscolor_details);
}


static void
gst_suppresscolor_class_init (GstSuppressColorClass * klass)
{
	GObjectClass *gobject_class;
	GstBaseTransformClass *trans_class;

	gobject_class = (GObjectClass *) klass;
	trans_class = (GstBaseTransformClass *) klass;

	gobject_class->set_property = gst_suppresscolor_set_property;
	gobject_class->get_property = gst_suppresscolor_get_property;

	g_object_class_install_property (gobject_class, PROP_SUPPRESS_Y,
	                                 g_param_spec_boolean ("suppress-y", "Suppress Y", "Wether the Y compononent should be suppressed",
	                                                       FALSE, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, PROP_SUPPRESS_U,
	                                 g_param_spec_boolean ("suppress-u", "Suppress U", "Wether the U compononent should be suppressed",
	                                                       FALSE, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, PROP_SUPPRESS_V,
	                                 g_param_spec_boolean ("suppress-v", "Suppress V", "Wether the V compononent should be suppressed",
	                                                       FALSE, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, PROP_SUPPRESSION_VALUE_Y,
	                                 g_param_spec_int ("value-y", "Value Y", "On suppression, the new constant value of the Y component",
	                                                   0, 255, DEFAULT_SUPPRESSION_VALUE_Y, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, PROP_SUPPRESSION_VALUE_U,
	                                 g_param_spec_int ("value-u", "Value U", "On suppression, the new constant value of the U component",
	                                                   0, 255, DEFAULT_SUPPRESSION_VALUE_U, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, PROP_SUPPRESSION_VALUE_V,
	                                 g_param_spec_int ("value-v", "Value V", "On suppression, the new constant value of the V component",
	                                                   0, 255, DEFAULT_SUPPRESSION_VALUE_V, G_PARAM_READWRITE));

	trans_class->set_caps = GST_DEBUG_FUNCPTR (gst_suppresscolor_set_caps);
	trans_class->transform_ip = GST_DEBUG_FUNCPTR(gst_suppresscolor_transform_ip);

}


static void
gst_suppresscolor_init (GstSuppressColor * filter,
                        GstSuppressColorClass * gclass)
{
	filter->suppress_y = FALSE;
	filter->suppress_u = FALSE;
	filter->suppress_v = FALSE;
	filter->value_y = DEFAULT_SUPPRESSION_VALUE_Y;
	filter->value_u = DEFAULT_SUPPRESSION_VALUE_U;
	filter->value_v = DEFAULT_SUPPRESSION_VALUE_V;

	if (!(filter->suppress_y || filter->suppress_u ||  filter->suppress_v))
		gst_base_transform_set_passthrough (GST_BASE_TRANSFORM (filter), TRUE);
	else
		gst_base_transform_set_passthrough (GST_BASE_TRANSFORM (filter), FALSE);
}


static void
gst_suppresscolor_set_property (GObject * object, guint prop_id,
                                const GValue * value, GParamSpec * pspec)
{
	GstSuppressColor *filter;

	g_return_if_fail (GST_IS_SUPPRESSCOLOR (object));
	filter = GST_SUPPRESSCOLOR (object);

	switch (prop_id) {
		case PROP_SUPPRESS_Y:
			filter->suppress_y = g_value_get_boolean (value);
			break;
		case PROP_SUPPRESS_U:
			filter->suppress_u = g_value_get_boolean (value);
			break;
		case PROP_SUPPRESS_V:
			filter->suppress_v = g_value_get_boolean (value);
			break;
		case PROP_SUPPRESSION_VALUE_Y:
			filter->value_y = g_value_get_int (value);
			break;
		case PROP_SUPPRESSION_VALUE_U:
			filter->value_u = g_value_get_int (value);
			break;
		case PROP_SUPPRESSION_VALUE_V:
			filter->value_v = g_value_get_int (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}

	if (!(filter->suppress_y || filter->suppress_u ||  filter->suppress_v))
		gst_base_transform_set_passthrough (GST_BASE_TRANSFORM (filter), TRUE);
	else
		gst_base_transform_set_passthrough (GST_BASE_TRANSFORM (filter), FALSE);
}


static void
gst_suppresscolor_get_property (GObject * object, guint prop_id,
                                GValue * value, GParamSpec * pspec)
{
	GstSuppressColor *filter;

	g_return_if_fail (GST_IS_SUPPRESSCOLOR (object));
	filter = GST_SUPPRESSCOLOR (object);

	switch (prop_id) {
		case PROP_SUPPRESS_Y:
			g_value_set_boolean (value, filter->suppress_y);
			break;
		case PROP_SUPPRESS_U:
			g_value_set_boolean (value, filter->suppress_u);
			break;
		case PROP_SUPPRESS_V:
			g_value_set_boolean (value, filter->suppress_v);
			break;
		case PROP_SUPPRESSION_VALUE_Y:
			g_value_set_int (value, filter->value_y);
			break;
		case PROP_SUPPRESSION_VALUE_U:
			g_value_set_int (value, filter->value_u);
			break;
		case PROP_SUPPRESSION_VALUE_V:
			g_value_set_int (value, filter->value_v);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}


static gboolean
gst_suppresscolor_set_caps (GstBaseTransform * base, GstCaps * incaps,
                            GstCaps * outcaps)
{
	GstSuppressColor *filter;
	GstStructure * s;
	gboolean res;

	filter = GST_SUPPRESSCOLOR(base);

	GST_DEBUG_OBJECT (filter,
	                  "set_caps: in %" GST_PTR_FORMAT " out %" GST_PTR_FORMAT, incaps, outcaps);

	s = gst_caps_get_structure (incaps, 0);
	res = gst_structure_get_int (s, "width", &filter->width);
	res &= gst_structure_get_int (s, "height", &filter->height);

	if (res)
		filter->size = GST_VIDEO_I420_SIZE(filter->width, filter->height);

	return res;
}


static GstFlowReturn
gst_suppresscolor_transform_ip(GstBaseTransform * base, GstBuffer * buf)
{
	GstSuppressColor *filter;
	guint8 *data;
	guint size, y_offset, u_offset, v_offset;

	if (gst_base_transform_is_passthrough (base))
		return GST_FLOW_OK;

	filter = GST_SUPPRESSCOLOR (base);

	data = GST_BUFFER_DATA (buf);
	size = GST_BUFFER_SIZE (buf);

	if (size != filter->size)
	{
		GST_ELEMENT_ERROR (filter, STREAM, FORMAT, (NULL),
		                   ("Invalid buffer size %d, expected %d",
		                    size, filter->size));
		return GST_FLOW_ERROR;
	}

	y_offset = GST_VIDEO_I420_Y_OFFSET(filter->width, filter->height);
	u_offset = GST_VIDEO_I420_U_OFFSET(filter->width, filter->height);
	v_offset = GST_VIDEO_I420_V_OFFSET(filter->width, filter->height);

	if (filter->suppress_y)
		memset(data + y_offset, filter->value_y, u_offset - y_offset);

	if (filter->suppress_u)
		memset(data + u_offset, filter->value_u, v_offset - u_offset);

	if (filter->suppress_v)
		memset(data + v_offset, filter->value_v, size - v_offset);

	return GST_FLOW_OK;
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
SuppressColor_init (GstPlugin * SuppressColor)
{
	/* debug category for fltering log messages
	 *
	 * exchange the string 'Template SuppressColor' with your description
	 */
	GST_DEBUG_CATEGORY_INIT (gst_suppresscolor_debug, "suppresscolor",
	                         0, "Template SuppressColor");

	return gst_element_register (SuppressColor, "suppresscolor", GST_RANK_NONE,
	                             GST_TYPE_SUPPRESSCOLOR);
}

/* gstreamer looks for this structure to register SuppressColors
 *
 * exchange the string 'Template SuppressColor' with your SuppressColor description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "suppresscolor",
    "Template suppresscolor",
    SuppressColor_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
	)
