/**
 * suppresscolor.vala
 * This file is part of PlayRaw.
 *
 * Copyright © 2008 Stian Selnes
 *
 *
 * PlayRaw is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PlayRaw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PlayRaw.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *     Stian Selnes <stian.selnes@gmail.com>
 */

using GLib;
using Gst;

namespace Gst {
	public class SuppressColor : Gst.VideoFilter {


		private int width;
		private int height;
		private int size;

		private DebugCategory debug;

		[Description(nick="suppress-y", blurb="Wether the Y compononent should be suppressed")]
		public bool suppress_y {
			get;
			set;
			default = false;
		}

		[Description(nick="suppress-u", blurb="Wether the U compononent should be suppressed")]
		public bool suppress_u {
			get;
			set;
			default = false;
		}

		[Description(nick="suppress-v", blurb="Wether the V compononent should be suppressed")]
		public bool suppress_v {
			get;
			set;
			default = false;
		}

		[Description(nick="value-y", blurb="On suppression, the new constant value of the Y component")]
		public int value_y {
			get;
			set;
			default = 127;
		}

		[Description(nick="value-u", blurb="On suppression, the new constant value of the U component")]
		public int value_u {
			get;
			set;
			default = 127;
		}

		[Description(nick="value-v", blurb="On suppression, the new constant value of the V component")]
		public int value_v {
			get;
			set;
			default = 127;
		}

		class construct {
			set_details_simple (
				"Suppress color component(s)",
				"Filter/Video",
				"Suppress given color component(s) in video stream",
				"Stian Selnes <<stian.selnes@gmail.com>>");

			var sink_template = new PadTemplate (
				"sink",
				PadDirection.SINK,
				PadPresence.ALWAYS,
				Caps.from_string (VideoCaps.YUV("I420")));
			add_pad_template (sink_template);

			var src_template = new PadTemplate (
				"src",
				PadDirection.SRC,
				PadPresence.ALWAYS,
				Caps.from_string (VideoCaps.YUV("I420")));
			add_pad_template (src_template);
		}

		/*
		construct {
			debug = new DebugCategory ();
			debug.init ("suppresscolor", 0, "Suppression of color component");
		}
		*/

		public override bool set_caps (Gst.Caps incaps, Gst.Caps outcaps) {
			Structure s = incaps.get_structure (0);
			bool res = s.get_int ("width", out width);
			res &= s.get_int ("height", out height);

			if (res)
				size = video_format_get_size (VideoFormat.I420, width, height);

			return res;
		}


		public override FlowReturn transform_ip (Buffer buf) {
			if (is_passthrough ()) {
				return FlowReturn.OK;
			}


			if (buf.size != size) {
				/*
				  GST_ELEMENT_ERROR (filter, STREAM, FORMAT, (NULL),
				  ("Invalid buffer size %d, expected %d",
				  size, filter->size));
				*/
				/* debug.error ("Invalid buffer size %d, expected %d", buf.size, size); */
				return FlowReturn.UNEXPECTED;
			}

			int offset_y = video_format_get_component_offset (VideoFormat.I420, 0, width, height);
			int offset_u = video_format_get_component_offset (VideoFormat.I420, 1, width, height);
			int offset_v = video_format_get_component_offset (VideoFormat.I420, 2, width, height);

			if (suppress_y)
				Memory.set ((void*) buf.data + offset_y, value_y, offset_u - offset_y);
			if (suppress_u)
				Memory.set ((void*) buf.data + offset_u , value_u, offset_v - offset_u);
			if (suppress_v)
				Memory.set ((void*) buf.data + offset_v, value_v, size - offset_v);

			return FlowReturn.OK;
		}
	}
}