/**
 * pyraw-ui.vala
 * This file is part of PlayRaw.
 *
 * Copyright © 2008 Stian Selnes
 *
 *
 * PlayRaw is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PlayRaw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PlayRaw.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *     Stian Selnes <stian.selnes@gmail.com>
 */


using Gtk;

public class PlayRawUI : GLib.Object {

	private Builder builder;
	private weak Window window;
	private weak DrawingArea video_area;
	private RawPlayer player;

	// UI options for player behavior
	private weak ToggleAction toggleaction_show_y;
	private weak ToggleAction toggleaction_show_u;
	private weak ToggleAction toggleaction_show_v;
	private weak Statusbar statusbar;
	private uint statusbar_context_id;
	private uint statusbar_message_id;

	private int current_frame;
	private int total_frames;

	public PlayRawUI (RawPlayer player) {
		this.player = player;
		// Cannot update state of player from the streaming thread. Wait until idle.
		player.file_loaded += (p) => { Idle.add (update_player_properties); };
		// Cannot update UI from the streaming thread. Wait until idle.
		player.frame_changed += (p, current, total) => {
			current_frame = current;
			total_frames = total;
			Idle.add (update_statusbar);
		};
	}


	public bool setup () {
		try {
			builder = new Builder ();
			// TODO: find a correct way to get directory and file
			var path_ui_file = Path.build_filename (Config.PACKAGE_DATADIR,
													"interface.ui");
			builder.add_from_file (path_ui_file);
			window = (Window) builder.get_object ("mainWindow");
			weak ToolButton toolbutton_play =  (ToolButton) builder.get_object ("toolbtn_play");
			weak ToolButton toolbutton_pause =  (ToolButton) builder.get_object ("toolbtn_pause");
			weak ToolButton toolbutton_stop =  (ToolButton) builder.get_object ("toolbtn_stop");
			weak ToolButton toolbutton_prev =  (ToolButton) builder.get_object ("toolbtn_prev");
			weak ToolButton toolbutton_next =  (ToolButton) builder.get_object ("toolbtn_next");
			weak Action action_open = (Action) builder.get_object ("menuitem_open");
			weak Action action_quit = (Action) builder.get_object ("menuitem_quit");
			weak Action action_about = (Action) builder.get_object ("menuitem_about");
			toggleaction_show_y = (ToggleAction) builder.get_object ("check_ycomp");
			toggleaction_show_u = (ToggleAction) builder.get_object ("check_ucomp");
			toggleaction_show_v = (ToggleAction) builder.get_object ("check_vcomp");

			video_area = (DrawingArea) builder.get_object ("da_video");
			statusbar = (Statusbar) builder.get_object ("statusbar");
			statusbar_context_id = statusbar.get_context_id ("main window");
			statusbar_message_id = statusbar.push (statusbar_context_id, "No video loaded");

			window.destroy += on_quit;
			toolbutton_play.clicked += on_play;
			toolbutton_pause.clicked += on_pause;
			toolbutton_stop.clicked += on_stop;
			toolbutton_prev.clicked += on_prev;
			toolbutton_next.clicked += on_next;
			action_open.activate += on_open;
			action_quit.activate += on_quit;
			action_about.activate += on_about;

			toggleaction_show_y.toggled += (ta) => { update_player_properties (); };
			toggleaction_show_u.toggled += (ta) => { update_player_properties (); };
			toggleaction_show_v.toggled += (ta) => { update_player_properties (); };

			window.show_all ();

		} catch (GLib.Error err) {
			var msg = new Gtk.MessageDialog (
				null, Gtk.DialogFlags.MODAL, Gtk.MessageType.ERROR, Gtk.ButtonsType.CANCEL,
				"Failed to load UI\n" + err.message);
			msg.run();
			return false;
		}
		return true;
	}


	public void open_file (string? filename) {
		OpenDialog dialog;
		if (filename == null)
			dialog = new OpenDialog (builder);
		else
			dialog = new OpenDialog.select_file (builder, filename);

		if (dialog.run ()) {
			video_area.set_size_request (dialog.input_width, dialog.input_height);
			player.load_file (dialog.input_filename,
			                  dialog.input_width,
			                  dialog.input_height,
			                  dialog.input_framerate,
			                  dialog.input_color_format,
			                  Gdk.x11_drawable_get_xid (video_area.window));
		}
	}


	private void on_play (ToolButton button) {
			player.play ();
	}


	private void on_pause (ToolButton button) {
			player.pause ();
	}


	private void on_stop (ToolButton button) {
			player.stop ();
	}


	private void on_prev (ToolButton button) {
			player.prev_frame ();
	}


	private void on_next (ToolButton button) {
			player.next_frame ();
	}


	private void on_open (Action item) {
		open_file (null);
	}


	private void on_about (Action item) {
		// TODO: Fix one entry for each author if multiple authors
		string[] authors = { Config.AUTHORS, null };
		string license_text =
		"%s is free software: you can redistribute it and/or modify\n" +
		"it under the terms of the GNU General Public License as published by\n" +
		"the Free Software Foundation, either version 3 of the License, or\n" +
		"(at your option) any later version.\n" +
		"\n" +
		"%s is distributed in the hope that it will be useful,\n" +
		"but WITHOUT ANY WARRANTY; without even the implied warranty of\n" +
		"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n" +
		"GNU General Public License for more details.\n" +
		"\n" +
		"You should have received a copy of the GNU General Public License\n" +
		"along with %s. If not, see <http://www.gnu.org/licenses/>.";
		license_text = license_text.printf (
			Config.PACKAGE_NAME, Config.PACKAGE_NAME, Config.PACKAGE_NAME);

		var about = new AboutDialog ();
		about.program_name = Config.PACKAGE_NAME;
		about.version = Config.PACKAGE_VERSION;
		try {
			about.logo = new Gdk.Pixbuf.from_file (
				Path.build_filename (Config.PACKAGE_DATADIR, "logo.svg"));
		} catch (GLib.Error e) {}
		about.authors = authors;
		about.copyright = "Copyright © 2008 Stian Selnes";
		about.license = license_text;
		about.comments = "Tool for viewing and analyzing raw video files";
		about.copyright = "Copyright © 2008 Stian Selnes";
		about.website = "http://stianse.wordpress.com/playraw/";
		// FIXME: The website-label does not seem to work
		about.website_label = "Visit the homepage";
		about.run ();
		about.destroy ();
	}


	private void on_quit () {
		player.close_file ();
		Gtk.main_quit ();
	}


	private bool update_statusbar () {
		statusbar.pop (statusbar_context_id);
		statusbar_message_id = statusbar.push (
			statusbar_context_id,
			"Frame %d/%d".printf (current_frame, total_frames-1));
		// remove from list of idle functions
		return false;

	}


	private bool update_player_properties () {
		if (toggleaction_show_y.active ||
		    toggleaction_show_u.active ||
		    toggleaction_show_v.active)
		{
			player.show_color_components (toggleaction_show_y.active,
			                              toggleaction_show_u.active,
			                              toggleaction_show_v.active);
		} else {
			player.show_color_components (true, true, true);
		}
		// returning "false" means remove from list of idle functions
		return false;
	}
}



public class OpenDialog : GLib.Object{

	public Builder builder              { get; construct; }

	public string input_filename        { get; private set; }
	public int input_width              { get; private set; }
	public int input_height             { get; private set; }
	public int input_framerate          { get; private set; }
	public int input_color_format       { get; private set; }

	private weak Dialog dialog;
	private weak FileChooserButton file_chooser_button;
	private weak Entry entry_width;
	private weak Entry entry_height;
	private weak Entry entry_fps;
	private weak ComboBox combobox_format;


	construct {
		dialog = (Dialog) builder.get_object ("openDialog");
		file_chooser_button = (FileChooserButton) builder.get_object ("filechooserbutton");
		entry_width = (Entry) builder.get_object ("entry_width");
		entry_height = (Entry) builder.get_object ("entry_height");
		entry_fps = (Entry) builder.get_object ("entry_fps");
		combobox_format = (ComboBox) builder.get_object ("combobox_format");

		combobox_format.set_active (0); // need to set this manually, bug?
		file_chooser_button.file_set += on_file_set;
	}


	public OpenDialog (Builder builder) {
		this.builder = builder;
		input_filename = null;
	}


	public OpenDialog.select_file (Builder builder, string filename) {
		this.builder = builder;
		input_filename = filename;
	}


	public bool run () {
		bool result = false;

		if (input_filename != null)
			file_chooser_button.set_filename (input_filename);

		if (dialog.run () == Gtk.ResponseType.OK) {
			// FIXME: Check for valid entries and loop if not valid
			input_width = entry_width.get_text ().to_int ();
			input_height = entry_height.get_text ().to_int ();
			input_framerate = entry_fps.get_text ().to_int ();
			var format_str = combobox_format.get_active_text ();
			// FIXME: Use enums and maybe index instead of text
			if (format_str == "I420")
				input_color_format = 0;
			else
				input_color_format = -1;
			result = true;
		}
		dialog.hide ();
		return result;
	}


	private void on_file_set (FileChooserButton chooser) {
		input_filename = chooser.get_filename ();
	}
}


public static int main (string[] args) {
	if (args.length > 1) {
		if (args[1] == "-h" || args[1] == "--help") {
			stdout.printf ("%s %s\n\n", Config.PACKAGE_NAME, Config.PACKAGE_VERSION);
			stdout.printf ("Usage:\n\t%s [filename]\n", args[0]);
			return 0; // drop this if we want to pass --help to gst.init
		}
	}

	Gtk.init (ref args);
	Gst.init (ref args);


	var player = new RawPlayer ();
	var mainapp = new PlayRawUI (player);
	if (mainapp.setup ()) {
		if (args.length > 1) {
			string filepath = args[1];
			if (!args[1].has_prefix("/")) {
				string cwd = Os.get_current_dir_name ();
				filepath = Path.build_filename (cwd, filepath);
			}
			mainapp.open_file (filepath);
		}
		Gtk.main ();
	}

	return 0;
}