/**
 * rawplayer.vala
 * This file is part of PlayRaw.
 *
 * Copyright © 2008 Stian Selnes
 *
 *
 * PlayRaw is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PlayRaw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PlayRaw.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *     Stian Selnes <stian.selnes@gmail.com>
 */

using Gst;


public class RawPlayer: GLib.Object {

	public signal void file_loaded ();
	public signal void file_closed ();
	public signal void frame_changed (int current_frame, int total_frames);


	// FIXME: Restrict access on these properties
	public string filename      {get; set;}
	public uint width           {get; set;}
	public uint height          {get; set;}
	public uint framerate       {get; set;}
	public uint format          {get; set;}
	public uint xid             {get; set;}

	private bool show_y = true;
	private bool show_u = true;
	private bool show_v = true;

	private Pipeline pipeline;
	private dynamic Element suppresscolor;
	private bool loading_complete;
	private int64 duration = -1;


	~RawPlayer () {
		close_file ();
	}

	public void load_file (string filename, uint width, uint height,
						   uint framerate, uint format, uint xid) {
		if (pipeline != null) {
			close_file ();
		}

		this.filename = filename;
		this.width = width;
		this.height = height;
		this.framerate = framerate;
		this.format = format;
		this.xid = xid;

		loading_complete = false;

		pipeline = new Pipeline ("RawPlayerPipeline");
		var bus = pipeline.get_bus ();
		bus.add_signal_watch ();
		bus.enable_sync_message_emission ();
		bus.message += on_message;
		bus.sync_message += on_sync_message;

		dynamic Element src = ElementFactory.make ("filesrc", null);
		src.location = filename;
		pipeline.add (src);

		dynamic Element videoparse = ElementFactory.make ("videoparse", null);
		videoparse.width = width;
		videoparse.height = height;
		videoparse.set ("framerate", framerate, 1);
		pipeline.add (videoparse);

		suppresscolor = ElementFactory.make ("suppresscolor", null);
		suppresscolor.suppress_y = ! show_y;
		suppresscolor.suppress_u = ! show_u;
		suppresscolor.suppress_v = ! show_v;
		pipeline.add (suppresscolor);

		Element colorspace = ElementFactory.make ("ffmpegcolorspace", null);
		pipeline.add (colorspace);

		dynamic Element sink = ElementFactory.make ("ximagesink", null);
		sink.force_aspect_ratio = true;
		pipeline.add (sink);

		src.link_many (videoparse, suppresscolor, colorspace, sink);

		var sink_pad = sink.get_static_pad ("sink");
		sink_pad.add_buffer_probe (on_sink_buffer);

		// Set state to playing so we push the first frame to the sink.
		// Will be paused after first frame is shown.
		pipeline.set_state (State.PLAYING);
	}


	public void close_file () {
		if (pipeline == null)
			return;
		pipeline.set_state (State.NULL);
		pipeline = null;
		loading_complete = false;
		file_closed ();
	}


	public void play () {
		pipeline.set_state (State.PLAYING);
	}


	public void pause () {
		pipeline.set_state (State.PAUSED);
	}


	public void stop () {
		// Pause and go to first frame
		pipeline.set_state (State.PAUSED);
		var format = Format.DEFAULT;
		if (!pipeline.seek_simple (format, SeekFlags.FLUSH, 0)) {
			debug ("Error while seeking for first frame");
		}
	}


	public void prev_frame () {
		// Pause and go to next frame
		pipeline.set_state (State.PAUSED);
		advance_frames (-1);
	}


	public void next_frame () {
		// Pause and go to next frame
		pipeline.set_state (State.PAUSED);
		advance_frames (1);
	}


	private void on_message (Bus bus, Message message) {
		switch (message.type) {

		case MessageType.ERROR:
			GLib.Error e;
			string dbg;
			message.parse_error (out e, out dbg);
			critical ("ERROR: %s", e.message);
			debug ("%s", dbg);
			break;

		case MessageType.EOS:
		default:
			break;
		}
	}


	private void on_sync_message (Bus bus, Message message) {

		if (message.structure != null &&
		    message.structure.has_name ("prepare-xwindow-id")) {
			((XOverlay) message.src).set_xwindow_id (xid);

			// Pause when we see the first frame. In "loaded" state.
			if (loading_complete == false) {
				pipeline.set_state (State.PAUSED);
			}
		}
		if (loading_complete == false
		    && message.src == pipeline
		    && message.structure != null
		    && message.structure.has_name ("GstMessageState")) {
			State old_state, new_state, pending_state;
			message.parse_state_changed (out old_state, out new_state, out pending_state);
			if (old_state == State.READY && new_state == State.PAUSED) {
				loading_complete = true;
				file_loaded ();
			}
		}
	}

	private void advance_frames (int n) {
		int64 pos;
		var format = Format.DEFAULT;

		if (pipeline.query_position (ref format, out pos)) {
			int64 seek_frame = pos + n;
			if (seek_frame < 0) {
				seek_frame = 0;
			} else if (seek_frame > duration - 1) {
				seek_frame = duration - 1;
			}
			if (!pipeline.seek_simple (format, SeekFlags.FLUSH, seek_frame)) {
				debug ("Error while seeking frame %d", (int) seek_frame);
			}
		} else {
			debug ("Error while quering current position");
		}
	}


	public void show_color_components (bool y, bool u, bool v) {
		show_y = y;
		show_u = u;
		show_v = v;
		if (pipeline != null)
			update_pipeline ();
	}


	private void update_pipeline ()
	requires (pipeline != null) {
		var state = pipeline.current_state;
		pipeline.set_state (State.PAUSED);
		suppresscolor.suppress_y = ! show_y;
		suppresscolor.suppress_u = ! show_u;
		suppresscolor.suppress_v = ! show_v;
		advance_frames (0);
		pipeline.set_state (state);
	}


	private bool on_sink_buffer (Pad pad, Buffer buffer) {
		var format = Format.DEFAULT;
		int64 pos;

		if (duration < 0) {
			if (pipeline.query_duration (ref format, out duration) == false) {
				debug ("Failed to query length of sequence");
			}
		}

		if (pipeline.query_position (ref format, out pos)) {
			frame_changed ((int) pos, (int) duration);
		} else {
			debug ("Error while quering position");
		}
		return true;
	}
}
